#! /usr/bin/perl
#! /usr/bin/perl -w
#
#-----------------------------------------------------------------------------
#   Open Source License/Disclaimer, 
#   Global Systems Division NOAA/OAR/ESRL/GSD/ITS
#   325 Broadway Boulder, CO 80305
#
#    This software is distributed under the Open Source Definition, 
#    which may be found at http://www.opensource.org/osd.html.
#
#    In particular, redistribution and use in source and binary forms, 
#    with or without modification, are permitted provided that the following 
#    conditions are met:
#
#      * Redistributions of source code must retain this notice, this list of 
#        conditions and the following disclaimer.
#      * Redistributions in binary form must provide access to this notice, this 
#        list of conditions and the following disclaimer, and the underlying 
#        source code.
#      * All modifications to this software must be clearly documented, and are 
#        solely the responsibility of the agent making the modifications.
#      * If significant modifications or enhancements are made to this software, 
#        the GSD Software Policy Manager (softwaremgr.gsd@noaa.gov) should be notified.
#
#    THIS SOFTWARE AND ITS DOCUMENTATION ARE IN THE PUBLIC DOMAIN AND ARE FURNISHED 
#    "AS IS."  THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS INSTRUMENTALITIES, 
#    OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY, EXPRESS OR IMPLIED, AS TO THE 
#    USEFULNESS OF THE SOFTWARE AND DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO 
#    RESPONSIBILITY (1) FOR THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO 
#    PROVIDE TECHNICAL SUPPORT TO USERS.
#
#
#  McIDASArea2NetCDF.pl - driver script for McArea2NetCDF java method
#
#  McArea2NetCDF.pl \
#       -logDir <logDir> \
#        -cfgFile <cfgFile>
#
#  where:
#
#  -logDir <logDir>  --  the directory into which a log file will be created
#       and log messages written to.
#  -cfgFile <cfgFile>  --  the configuration file containing information about
#       which files to get from where.
#  -incDir <includeDir>  --  a directory to prepend to @INC.  ftpDD relies
#       on the logTools.pl and copyNRun.pl Perl modules.  If not specified,
#       '/usr/local/rtoper/scripts' is used.
#
#  11/18/08  Bob Lipschutz 	Original version
#  09/28/12  Bob                Bumped java max memory setting (Xmx) from 512m to 640m
#  08/19/14  Bob                Added lock mechanism to avoid concurrent jobs
#  03/21/14  Bob                Updated for java netcdf 4.6.4 and separate visad packages
#-----------------------------------------------------------------------------
#  $Id: McIDASArea2NetCDF.pl,v 1.12 2016/03/21 23:58:02 lipschut Exp $

$0 =~ s|.*/([^/]+)$|$1|;

$usage = "$0 -cfgFile <cfgFile> [-logDir <logDir>] [-backPosition <back 1-n>] [-r(eplace)] [-proc <procName>] [-incDirs <inc>] [-v(erbose)]";

$LCL_DIR_IND = 1;       # index of the local directory value

$CHANNEL_IND  = 0;       # index of the channel value
$XMLPATH_IND  = 1;       # index of the xml file value
$OUTPATH_IND  = 2;       # index of the destination data directory value
$KEY_PRF_IND  = 3;       # index of the product key prefix value
$AREAFILE_IND = 4;       # index of the area file value


$verboseArg = "";    # No verbosity by default
$logArg = "";

$processName = $0;  	 # default process name

$McArea2NetCDF = "gov.noaa.esrl.mcidas.McArea2NetCDF";

$jarPath = "$ENV{'HOME'}/bin";
$classPath = "$jarPath/java-getopt.jar:$jarPath:log4j.jar:$jarPath/slf4j-jdk14-1.7.19.jar:$jarPath/netcdfAll-4.6.4.jar:$jarPath/visad.jar:$jarPath/HTTPClient.jar:$jarPath/McIDASTools.jar";

#
# Process the command-line arguments.
#
$logDir = "";
$daysBack = 0;
@incDirs = @pathDirs = ( );
for ($i = 0;  $i < @ARGV;  ++$i)
    {
    $lcArg = lc($ARGV[$i]);
    $lcArgLen = length($lcArg);
    if (substr('-logdirectory', 0, $lcArgLen) eq $lcArg)
        {
        $logDir = $ARGV[++$i];
	$logArg = "-l $logDir";
        }
    elsif (substr('-cfgfile', 0, $lcArgLen) eq $lcArg)
        {
        $cfgFile = $ARGV[++$i];
        }
    elsif (substr('-backposition', 0, $lcArgLen) eq $lcArg)
        {
        $backArg = "&pos=-$ARGV[++$i]";
        }
    elsif (substr('-replace', 0, $lcArgLen) eq $lcArg)
        {
        $replaceArg = "-r";
        }
    elsif (substr('-processname', 0, $lcArgLen) eq $lcArg)
        {
        $processName = $ARGV[++$i];
        }
    elsif ((substr('-incdirectory', 0, $lcArgLen) eq $lcArg)
      || (substr('-includedirectory', 0, $lcArgLen) eq $lcArg))
        {
        push(@incDirs, $ARGV[++$i]);
        }
    elsif (substr('-pathdirectory', 0, $lcArgLen) eq $lcArg)
        {
        push(@pathDirs, $ARGV[++$i]);
        }
    elsif (substr('-verbose', 0, $lcArgLen) eq $lcArg)
        {
        $verboseArg = "-v";
        }
    elsif (substr('-help', 0, $lcArgLen) eq $lcArg)
        {
        print STDERR "$usage\n";
        }
    else
        {
        print STDERR "$usage\n \
          $0:  ignoring unrecognized argument '$ARGV[$i]'.\n";
        }
    }
unless (defined($cfgFile))
    {
    print STDERR "$0:  no configuration file specified.  Aborting.\n";
    exit(1);
    }

@incDirs = ("$ENV{'HOME'}/bin", "$ENV{'HOME'}/scripts", '/usr/local/rtsys/scripts') unless (@incDirs > 0);
unshift(@INC, @incDirs);
@pathDirs = ("$ENV{'HOME'}/scripts", "$ENV{'HOME'}/scripts", '/usr/local/bin' )
    unless (@pathDirs > 0);
$ENV{'PATH'} = join(':', @pathDirs, $ENV{'PATH'});

require 'logTools.pl';
require 'lockTools.pl';

&LogFileOpen($logDir) if ($logDir ne '');

#------------------------------------------------------------------
# Set the lock file - allow only one instance to run at a time
#------------------------------------------------------------------
($cfg = $cfgFile)  =~ s|.*/([^/]+)$|$1|;

$lockDir = "/data/lock";
$lockDir = "/tmp" if (! -e $lockDir);
$lockFile = "$lockDir/$cfg.lock"; 
&LogMessages(" Setting lock file $lockFile ");
open  LOCK, ">>$lockFile" or die "Cannot open lock file $lockFile: $!";
$haveLock = &LockFile(\*LOCK, 1, 10);
die "Can't get lock for $lockFile (job must already be running), so exiting..." if $haveLock != 1;

truncate(LOCK,0);
print LOCK "$datetime | $h | $ENV{'JOB_ID'} | $0 @ARGV\n";

$start_time = time();

#
# Read the configuration file.
#
# Note:  much more documentation is found in the configuration file.
#
unless (open(CFG, $cfgFile))
    {
    &LogMessages(
      "! Could not open the configuration file '$cfgFile':  $!.  Aborting.");
    exit(1);
    }
@fileDefs = ( );
while (<CFG>)
    {
    next unless (/^\s*(\w+)\s*=\s*(.*)/);

    $lcName = lc($1);
    $lcNameLen = length($lcName);
    if (substr('classpath', 0, $lcNameLen) eq $lcName)
        {
        $classPath = $2;
        }
    elsif (substr('channel', 0, $lcNameLen) eq $lcName)
        {
        $currFileDefRef = [ ];
        push(@fileDefs, $currFileDefRef);
        @$currFileDefRef[$CHANNEL_IND] = $2;
        }
    elsif (substr('xmlpath', 0, $lcNameLen) eq $lcName)
        {
        @$currFileDefRef[$XMLPATH_IND] = $2;
        }
    elsif (substr('outpath', 0, $lcNameLen) eq $lcName)
        {
        @$currFileDefRef[$OUTPATH_IND] = $2;
        }
    elsif (substr('keyprefix', 0, $lcNameLen) eq $lcName)
        {
        @$currFileDefRef[$KEY_PRF_IND] = $2;
        }
    elsif (substr('areafile', 0, $lcNameLen) eq $lcName)
        {
        @$currFileDefRef[$AREAFILE_IND] = $2;
        }
    }
close(CFG);


#
# For each file grouping, process the specified Areafile
#
for ($i = 0;  $i < @fileDefs;  ++$i)
    {
    @fileDef = @{$fileDefs[$i]};

    $channel   = @fileDef[$CHANNEL_IND];
    $xmlPath   = @fileDef[$XMLPATH_IND];
    $outPath   = @fileDef[$OUTPATH_IND];
    $keyPrefix = @fileDef[$KEY_PRF_IND];
    $areaFile  = @fileDef[$AREAFILE_IND];

    $ichan = $i+1;
    &LogMessages("====== Starting channel #$ichan =======");
    &LogMessages("Channel:    $channel");
    &LogMessages("XML Path:   $xmlPath");
    &LogMessages("Out Path:   $outPath");
    &LogMessages("Key Prefix: $keyPrefix");
    &LogMessages("Area file:  $areaFile");

    $ext = "_$channel";
    $keyString = ".$keyPrefix" if ($keyPrefix);
    $proc = "$processName$keyString.$channel";


    $cmd = "java -Xmx640m -cp $classPath $McArea2NetCDF -a \"$areaFile$backArg\" -o $outPath -x $xmlPath -e $ext $logArg -p $proc $replaceArg $verboseArg";
    print "$cmd\n";

    $retcode = system($cmd) >> 8;
    &LogMessages("! $0 failed with exit code $retcode.")
       unless ($retcode == 0);

    }


$end_time = time;
$diff_time = $end_time - $start_time;
print "\nprocessing time = $diff_time sec\n";
printf "%s: completed  --  %s GMT  --\n", $0, scalar(gmtime($end_time));

printf LOCK "%s: completed in %d sec -- %s GMT --\n", $0, $diff_time, scalar(gmtime($end_time));
&UnlockFile(\*LOCK);
close LOCK;

exit 0;
