package gov.noaa.esrl.mcidas;

import java.util.Iterator;
import org.apache.log4j.Logger;

import edu.wisc.ssec.mcidas.AREAnav;
import edu.wisc.ssec.mcidas.AreaFileException;

import ucar.nc2.Variable;

public abstract class AreaDataWriter
    {
    protected Logger logger = null;
    protected boolean verbose = false;

    protected int imageTime;
    protected int dataWidth;
    protected int numBands;
    protected int lineRes_area = 1;
    protected int elemRes_area = 1;
    protected int lineStep = 1;
    protected int elemStep = 1;

    protected int startLine_area;
    protected int startElement_area;
    protected int numLines_area;
    protected int numElements_area;
   

    protected int startLine_nc = 0;
    protected int startElement_nc = 0;
    protected int[] startLine_gvar = {0,0,0,0,0,0,0,0,0};
    protected int[] startElement_gvar = {0,0,0,0,0,0,0,0,0};
    
    protected int lineRes_nc;
    protected int elemRes_nc;
    protected int numLines_nc;
    protected int numElements_nc;
    protected boolean haveLatLons;
    protected String latitude;
    protected String longitude;

    protected int[][][] area_data;
    protected byte[][] line_prefix;
    protected int[] cal;   
    protected int[] navArray;
    protected AREAnav  nav;
    protected int[] bands;

    protected AreaFileReader afr;
    protected NetCDFHandler nh;
    
    public AreaDataWriter(AreaFileReader afr, NetCDFHandler nh) throws AreaFileException
        {
        this.afr = afr;
        this.nh = nh;
        
        logger = MyLogger.logger;
        verbose = MyLogger.verbose;

        // Do some generic stuff
        loadAreaParameters();
        writeMcNameMapVars();
        
        area_data = afr.getAreaData();
        navArray = afr.getAreaNavArray();
        nav = afr.getAreaNavigation();  
        cal = afr.getCalArray();
        
        numLines_nc = nh.numLines_nc;
        numElements_nc = nh.numElements_nc;
        lineRes_nc = nh.lineRes_nc;
        elemRes_nc = nh.elemRes_nc;

        startLine_nc = nh.startLine_nc;
        startElement_nc = nh.startElement_nc;

        for (int i=0; i<nh.startLine_gvar.length; i++)
            {
            startLine_gvar[i] = nh.startLine_gvar[i];
            startElement_gvar[i] = nh.startElement_gvar[i];    
            }
        
        haveLatLons = nh.haveLatLons();
        latitude = nh.latName;
        longitude = nh.lonName;
        }
    
    public abstract void writeData();
    public abstract void writeNavAndCalVars();
   
    public void loadAreaParameters()
        {
        imageTime = afr.getStartTime();
        elemRes_area = (Integer) afr.adMap.get("element_resolution");
        lineRes_area = (Integer) afr.adMap.get("line_resolution");
        startLine_area = (Integer) afr.adMap.get("start_line");
        startElement_area = (Integer) afr.adMap.get("start_element");
        dataWidth = (Integer) afr.adMap.get("data_width");
        numElements_area = (Integer) afr.adMap.get("num_elements");
        numLines_area = (Integer) afr.adMap.get("num_lines");
        numBands = (Integer) afr.adMap.get("num_bands");
        bands = (int[]) afr.adMap.get("bands");
        }
    
    public void dumpAreaParameters()
        {
        logger.info("Image time:  " + imageTime);
        logger.info("Num lines:   " + numLines_area);
        logger.info("Num els:     " + numElements_area);
        logger.info("Start Line:  " + startLine_area);
        logger.info("Start Elem:  " + startElement_area);
        logger.info("Num bands:   " + numBands);
        logger.info("Sensor ID:   " + (Integer) afr.adMap.get("sensor_id"));
        logger.info("Sensor Type: " + (String) afr.adMap.get("sensor_type"));
        logger.info("Source Type: " + (String) afr.adMap.get("source_type"));
        logger.info("Data width:  " + dataWidth);
        logger.info("Element res: " + elemRes_area);
        logger.info("Line res:    " + lineRes_area);
        }
    
    public void writeMcNameMapVars()
        {
        Iterator iterator = nh.mcNameMap.keySet().iterator();
        String mcName;
        Variable v;

        // For each variable with a mcidas_name, write data value into netCDF
        while (iterator.hasNext())
            {
            mcName = (String) iterator.next();           
            if (afr.adMap.containsKey(mcName))
                {              
                v = nh.mcNameMap.get(mcName);               
                nh.writeVariableData(v.getName(), afr.adMap.get(mcName));
                }           
            }
        }
    
    public void writeStandardVars()
        {
        // Fill in some standard variables
        nh.writeVariableData("reference_time", (Object) new Double(imageTime)
                .doubleValue());
        nh.writeVariableData("num_lines", (Object) numLines_nc);
        nh.writeVariableData("num_elements", (Object) numElements_nc);
        nh.writeVariableData("ny", (Object) numLines_nc);
        nh.writeVariableData("nx", (Object) numElements_nc);
        nh.writeVariableData("data_width", (Object) dataWidth);
        nh.writeVariableData("line_resolution", (Object) lineRes_nc);
        nh.writeVariableData("element_resolution", (Object) elemRes_nc);
        nh.writeVariableData("line_step", (Object) lineStep);
        nh.writeVariableData("element_step", (Object) elemStep);
        nh.writeVariableData("start_line", (Object) startLine_nc);
        nh.writeVariableData("start_element", (Object) startElement_nc);

        int endLine = startLine_nc + numLines_nc * lineRes_nc;
        int endElem = startElement_nc + numElements_nc * elemRes_nc;
        nh.writeVariableData("end_line", (Object) endLine);
        nh.writeVariableData("end_element", (Object) endElem);
        }
    }
