package gov.noaa.esrl.mcidas;

import java.util.Date;
import java.util.Iterator;
import java.util.TreeMap;

import org.apache.log4j.Logger;

import edu.wisc.ssec.mcidas.AREAnav;
import edu.wisc.ssec.mcidas.AreaDirectory;
import edu.wisc.ssec.mcidas.AreaFileException;

public class AreaFileReader
    {
    private Logger logger = null;
    private boolean verbose = false;

    private AreaFile _af = null;
    private AreaDirectory _ad;

    public TreeMap<String, Object> adMap;

    private int[] bands = new int[20];

    public AreaFileReader(String areaFile) throws McAreaException
        {
        logger = MyLogger.logger;
        verbose = MyLogger.verbose;
        
        try
            {
            _af = new AreaFile(areaFile);
            _ad = _af.getAreaDirectory();
            
            loadAreaDirectoryMap();      
            }
        catch (AreaFileException afe)
            {
            throw new McAreaException("can't open Area File " + areaFile);
            }                  
        }

   
    /**
     * Generate a map containing name-value pairs for all the Area Directory
     * parameters.
     * @throws AreaFileException
     */
    private void loadAreaDirectoryMap() throws AreaFileException
        {
        adMap = new TreeMap<String, Object>();
        adMap.put("center_latitude", _ad.getCenterLatitude());
        adMap.put("center_latitude_res", _ad.getCenterLatitudeResolution());
        adMap.put("center_longitude", _ad.getCenterLongitude());
        adMap.put("center_longitude_res", _ad.getCenterLongitudeResolution());
        adMap.put("num_lines", _ad.getLines());
        adMap.put("num_elements", _ad.getElements());
        adMap.put("memo_field", _ad.getMemoField());
        adMap.put("nominal_time", _ad.getNominalTime());
        adMap.put("sensor_id", _ad.getSensorID());
        adMap.put("sensor_type", _ad.getSensorType());
        adMap.put("source_type", _ad.getSourceType());
        adMap.put("start_time", _ad.getStartTime());
        adMap.put("calibration_type", _ad.getCalibrationType());
        adMap.put("calibration_unit_name", _ad.getCalibrationUnitName());
        adMap.put("calibration_scale_factor", _ad.getCalibrationScaleFactor());
        
        adMap.put("data_width", _ad.getValue(AreaFile.AD_DATAWIDTH));
        adMap.put("data_offset", _ad.getValue(AreaFile.AD_DATAOFFSET));
        adMap.put("doc_length", _ad.getValue(AreaFile.AD_DOCLENGTH));
        adMap.put("cal_length", _ad.getValue(AreaFile.AD_CALLENGTH));
        adMap.put("cal_offset", _ad.getValue(AreaFile.AD_CALOFFSET));
        adMap.put("element_resolution", _ad.getValue(AreaFile.AD_ELEMRES));
        adMap.put("level_length", _ad.getValue(AreaFile.AD_LEVLENGTH));
        adMap.put("line_resolution", _ad.getValue(AreaFile.AD_LINERES));
        adMap.put("nav_offset", _ad.getValue(AreaFile.AD_NAVOFFSET));
        adMap.put("num_bands", _ad.getValue(AreaFile.AD_NUMBANDS));
        adMap.put("prefix_size", _ad.getValue(AreaFile.AD_PFXSIZE));
        adMap.put("project_number", _ad.getValue(AreaFile.AD_PROJNUM));
        adMap.put("status", _ad.getValue(AreaFile.AD_STATUS));
        adMap.put("start_line", _ad.getValue(AreaFile.AD_STLINE));
        adMap.put("start_element", _ad.getValue(AreaFile.AD_STELEM));
        adMap.put("val_code", _ad.getValue(AreaFile.AD_VALCODE));
        adMap.put("version", _ad.getValue(AreaFile.AD_VERSION));

        int date = _ad.getValue(AreaFile.AD_CRDATE);
        if (date < 200000) date += 1900000;
        adMap.put("creation_date", date);
        adMap.put("creation_time", _ad.getValue(AreaFile.AD_CRTIME));

        date = _ad.getValue(AreaFile.AD_IMGDATE);
        if (date < 200000) date += 1900000;
        adMap.put("image_date", date);
        adMap.put("image_time", _ad.getValue(AreaFile.AD_IMGTIME));

        bands = _ad.getBands();
        adMap.put("bands", bands);
        adMap.put("band", new Byte((byte) bands[0]));  // Save the 1st band... only have 1 for GVAR images

        int satId = _ad.getSensorID();
        if (satId >= 70 && satId <= 79)
            satId = satId / 2 - 27; // eg., GOES-8 through GOES-12 -> g12 = 12
        else if (satId >= 180 && satId <= 187) // GOES-13 through -16** (** speculation)
            satId = satId / 2 - 77;
        adMap.put("satellite_number", satId);
        }
    
    public int[][][] getAreaData() throws AreaFileException
        {
        return  _af.getData();
        }

    public AREAnav getAreaNavigation() throws AreaFileException
        {
        return _af.getNavigation();
        }

    public int[] getAreaNavArray() throws AreaFileException
        {
        return _af.getNav();
        }
 
    public int[] getCalArray() throws AreaFileException
        {
        return _af.getCal();
        }
    
   public int getStartTime()
        {
        Date date = (Date) adMap.get("start_time");
        long imageTime = date.getTime() / 1000;
        return (int) imageTime;
        }

    public void dumpAreaDirectoryMap()
        {
        Iterator iterator = adMap.keySet().iterator();
        String s;
        System.out.println("=== Area Directory data from map: ");
        while (iterator.hasNext())
            {
            s = (String) iterator.next();
            System.out.println(s + ":  " + adMap.get(s));
            }
        
        for (int i = 0; i < (Integer) adMap.get("num_bands"); i++)
            System.out.println(" band:  " + bands[i]);
        
        System.out.println("===");
        } 
    
    public static void listMcIDASNames()
        {
        System.out.print("\nmcidas_names:\n\n " +
                "center_latitude\n " +
                "center_longitude\n " +
                "num_lines\n " +
                "num_elements\n " +
                "memo_field\n " +
                "nominal_time\n " +
                "sensor_id\n " +
                "sensor_type\n " +
                "source_type\n " +
                "start_time\n " +
                "calibration_type\n " +
                "calibration_unit_name\n " +
                "calibration_scale_factor\n " +
                "data_width\n " +
                "data_offset\n " +
                "doc_length\n " +
                "element_resolution\n " +
                "level_length\n " +
                "line_resolution\n " +
                "nav_offset\n " +
                "num_bands\n " +
                "prefix_size\n " +
                "project_number\n " +
                "status\n " +
                "start_line\n " +
                "start_element\n " +
                "val_code\n " +
                "version\n " +
                "creation_date\n " +
                "creation_time\n " +
                
                "image_date\n " +                                                                                                                                                     
                "image_time\n " +                                                                                                                                                         
                "bands\n " +                                                                                                                                                         
                "band\n "                                                                                                                                                          
                );            
        }
    }


