package gov.noaa.esrl.mcidas;

import edu.wisc.ssec.mcidas.AreaFileException;
import ucar.ma2.Array;
import ucar.ma2.ArrayDouble;
import ucar.ma2.ArrayFloat;
import ucar.ma2.ArrayInt;
import ucar.ma2.Index;

public class AreaGvarImagerWriter extends AreaDataWriter
    {

    public AreaGvarImagerWriter(AreaFileReader afr, NetCDFHandler nh) throws AreaFileException
        {
        super(afr, nh);
        }
    
    public void writeData()
        {
        int[] imageOrigin = new int[2];
        ArrayInt.D2 imageLine_nc = new ArrayInt.D2(numBands, numElements_nc);
        
        int[] llOrigin = new int[2];
        llOrigin[0] = 0;
        llOrigin[1] = 0;
        float linEles[][] = new float[2][1];
        float latLons[][] = new float[2][1];
        Array aLats = new ArrayFloat.D2(1, numElements_nc);
        Array aLons = new ArrayFloat.D2(1, numElements_nc);
        Index iLats = aLats.getIndex();
        Index iLons = aLons.getIndex();        
        
        int pixelValue;
        int shift = 0;
        int iLine_area;
        int iElem_area;
        int iStartLine_area = 0;
        int iStartElem_area = 0;
                
        // Need to shift 5 bits for Raw GVAR... see Imager Data Block info under
        // http://www.unidata.ucar.edu/software/mcidas/miscdocs/area2.html
        // Also, may need to get startLine/El_nc from satellite-specific gvar coords
        shift = 5;
            
            // Starting coords are satellite-specific, and may be specified for each sat
        if (startLine_nc == 0)
            {
            int satIndex = (Integer) afr.adMap.get("satellite_number") - 8;
            startLine_nc = startLine_gvar[satIndex];
            startElement_nc = startElement_gvar[satIndex];
                       
            logger.info("Using starting coords for GOES-" + (satIndex + 8));
            logger.info("Starting line_nc/el_nc " + startLine_nc + "  " + startElement_nc);
            }

        if (startLine_nc != 0) 
            iStartLine_area = Math.round((startLine_nc - startLine_area) / (float)lineRes_area);
        if (startElement_nc != 0) 
            iStartElem_area = Math.round((startElement_nc - startElement_area) / (float)elemRes_area);

        if (lineRes_nc != lineRes_area) lineStep = lineRes_nc / lineRes_area;
        if (elemRes_nc != elemRes_area) elemStep = elemRes_nc / elemRes_area;
        if (lineStep == 0) lineStep = 1;
        if (elemStep == 0) elemStep = 1;
        iLine_area = iStartLine_area - lineStep;
        
        logger.info("in AreaGvarImageWriter: line,elemRes_nc" + "  " + lineRes_nc + "  " + elemRes_nc);
        logger.info("start_line_area, numLines: " + iStartLine_area  + "  " + numLines_nc);
        logger.info("start_elem_area, numElems: " + iStartElem_area  + "  " + numElements_nc);      
        logger.info("lineStep, elemStep (rel to area): " + lineStep  + "  " + elemStep);    
        logger.info("write latlons:  " + haveLatLons);
        
        try
            {            
            // Loop through lines in netCDF file
            for (int iLine_nc = 0; iLine_nc < numLines_nc; iLine_nc++)
                {                
                iLine_area += lineStep;
                
                // Fill with 'missing' if line is outside the Area
                if (iLine_area < 0 || iLine_area >= numLines_area ) 
                    {
                    for (int iElem_nc = 0; iElem_nc < numElements_nc; iElem_nc++)
                        {
                        latLons[0][0] = Float.NaN;
                        latLons[1][0] = Float.NaN;   
                        if (haveLatLons)
                            {
                            aLats.setFloat(iLats.set(0, iElem_nc), (float) latLons[0][0]);
                            aLons.setFloat(iLons.set(0, iElem_nc), (float) latLons[1][0]);
                            }
                        
                        // Assume data width = 2 (short)
                        imageLine_nc.set(0, iElem_nc, (short) 0);
                        }
                    }               
                else   // Line is within the Area
                    {
                    iElem_area = iStartElem_area - elemStep;
                    linEles[1][0] = iLine_area;

                    // Loop through elements in netCDF file
                    for (int iElem_nc = 0; iElem_nc < numElements_nc; iElem_nc++)
                        {
                        iElem_area += elemStep;

                        // Fill with 'missing' if element is outside the Area
                        if (iElem_area < 0 || iElem_area >= numElements_area)
                            {
                            pixelValue = 0;
                            latLons[0][0] = Float.NaN;
                            latLons[1][0] = Float.NaN;
                            }
                        else      // Element is within the Area
                            {
                            pixelValue = area_data[0][iLine_area][iElem_area];
                            if (haveLatLons)
                                {
                                linEles[0][0] = iElem_area;
                                latLons = nav.toLatLon(linEles);
                                }
                            }

                        if (haveLatLons)
                            {
                            aLats.setFloat(iLats.set(0, iElem_nc),
                                    (float) latLons[0][0]);
                            aLons.setFloat(iLons.set(0, iElem_nc),
                                    (float) latLons[1][0]);
                            }

                        if (dataWidth == 1)
                            imageLine_nc.set(0, iElem_nc, (byte) pixelValue);
                        else if (dataWidth == 2)
                            {
                            // Need to shift GVAR
                            pixelValue = (pixelValue >> shift);
                            imageLine_nc.set(0, iElem_nc, (short) pixelValue);
                            }
                       
/*                          if (iElem_nc >= 500 && iElem_nc < 530 && iLine_nc >=
                          350 && iLine_nc < 380) System.out.println("nc: " +
                          iLine_nc + " " + iElem_nc + " area: " + iLine_area + " " +
                          iElem_area + " " + " swapped: " + pixelValue + " " +
                          Integer.toHexString(pixelValue) + " Lat/lon: " +
                          latLons[0][0] + " " + latLons[1][0]);*/
                       
                        }
                    }
                
                // Write the line's data into the netCDF file
                imageOrigin[0] = iLine_nc;
                imageOrigin[1] = 0;
                nh.writeArrayData("image", imageOrigin, (Array) imageLine_nc);

                if (haveLatLons)
                    {
                    llOrigin[0] = iLine_nc;
                    nh.writeArrayData(latitude, llOrigin, (Array) aLats);
                    nh.writeArrayData(longitude, llOrigin, (Array) aLons);
                    }
                }
            }
        catch (Exception e)
            {
            logger.error("Can't write image data: ", e); 
            }
        }
    
    public void writeNavAndCalVars()
        {
        String[] wavelengths = {"vis", "4u ", "wv ", "11u", "12u", "13u"};
        String[] imcIds = {"AAAA", "BBBB", "I001", "W001", "S002", "N000", "XXXX", "YYYY", "ZZZZ"}; // for GOES-8, -9, ...
 
        final float rad2deg            = 180.0f/3.141592654f;
        final float app_earth_diameter = 2.0232f;
        final float ew_diameter        = 6378.160f;
        final float ns_diameter        = 6356.775f;
        final float Nr                 = 6.6166f; 

        int satId = (Integer) afr.adMap.get("satellite_number");
        
        //-------------------------------------------------------------------------
        // Fix up x,y-step to be relative to native resolution
        //-------------------------------------------------------------------------
        int band = bands[0];   // We only have one GVAR Imager band
        
        if (band == 1)         // Visible is 1-km res
            {
            nh.writeVariableData("line_step", lineRes_nc);
            nh.writeVariableData("element_step", elemRes_nc);
            }
        else if (band == 3)    // WV
            {
            if (satId < 12)    // 8-km WV line res before GOES-12
                nh.writeVariableData("line_step", lineRes_nc/8);  
            else               // 4-km WV line res GOES-12 & after
                nh.writeVariableData("line_step", lineRes_nc/4);  
            nh.writeVariableData("element_step", elemRes_nc/4);
            }
        else                   // 4-km res for other IR bands
            {
            nh.writeVariableData("line_step", lineRes_nc/4);
            nh.writeVariableData("element_step", elemRes_nc/4);
            }

        //-------------------------------------------------------------------------
        // Navigation parameters
        // For GVAR nav block, see 
        // http://www.ssec.wisc.edu/mcidas/doc/prog_man/2003/formats-4.html
        //-------------------------------------------------------------------------
       
        nh.writeVariableData("scan_status", navArray[2]);
        byte imc = 1;
        if ( (navArray[2] & (1 << 8)) != 0 ) imc = 0;
        nh.writeVariableData("imc_flag", imc);
        if (imc == 1 && satId <= 16)          
            nh.writeVariableData("imc_id", imcIds[satId-8]);           
        else
            nh.writeVariableData("imc_id", "    "); 
        
        byte iflip = 0;
        if ( (navArray[3] & (1 << 16)) != 0) iflip = 1;
        nh.writeVariableData("yaw_flip_flag", iflip);
        
        nh.writeVariableData("reference_longitude", navArray[5]/10000000.0f);
        nh.writeVariableData("reference_distance", navArray[6]/10000000.0f);
        nh.writeVariableData("reference_latitude", navArray[7]/10000000.0f);
        nh.writeVariableData("reference_orbit_yaw", navArray[8]/10000000.0f);
        nh.writeVariableData("reference_attitude_roll", navArray[9]/10000000.0f);
        nh.writeVariableData("reference_attitude_pitch", navArray[10]/10000000.0f);
        nh.writeVariableData("reference_attitude_yaw", navArray[11]/10000000.0f);
        nh.writeVariableData("imc_enable_time", navArray[14]/100.0f);
        nh.writeVariableData("spacecraft_comp_roll", navArray[15]/100.0f);
        nh.writeVariableData("spacecraft_comp_pitch", navArray[16]/100.0f);
        nh.writeVariableData("spacecraft_comp_yaw", navArray[16]/100.0f);       
        nh.writeVariableData("east_west_cycles", new Integer(navArray[380]).shortValue());
        nh.writeVariableData("east_west_increments", new Integer(navArray[382]).shortValue());
        nh.writeVariableData("north_south_cycles", new Integer(navArray[379]).shortValue());
        nh.writeVariableData("north_south_increments", new Integer(navArray[381]).shortValue());
        nh.writeVariableData("wavelength", wavelengths[band-1]);

        //-------------------------------------------------------------------------
        // Derived variables
        //-------------------------------------------------------------------------
        float Lop = (navArray[5]/10000000.0f) * rad2deg;
        float Lap = (navArray[7]/10000000.0f) * rad2deg;
        nh.writeVariableData("Lop", Lop);
        nh.writeVariableData("Lap", Lap);
        
        float dx = (app_earth_diameter * ew_diameter)/elemRes_nc;
        float dy = (app_earth_diameter * ns_diameter)/lineRes_nc;
        nh.writeVariableData("dx", dx);
        nh.writeVariableData("dy", dy);
        
        // Subpoint line/pixel are not available in Area File, 
        // so calculate from subpoint (reference) lat/lon
        float linEle[][] = new float[2][1];
        float latLon[][] = new float[2][1];
        latLon[nav.indexLat][0] = Lap;
        latLon[nav.indexLon][0] = Lop;
        linEle = nav.toLinEle(latLon);
        float xp = linEle[nav.indexEle][0] * elemRes_area + startElement_area; 
        float yp = linEle[nav.indexLine][0] * lineRes_area + startLine_area;
        nh.writeVariableData("Xp", new Float(xp).shortValue());
        nh.writeVariableData("Yp", new Float(yp).shortValue());    
        
        float orientation = 0.f;   // not sure how to derive this from Area file
        nh.writeVariableData("orientation", orientation);
        nh.writeVariableData("Nr", Nr);
        
        float Rx = new Float((2.f * Math.asin(1.d/Nr))/dx).floatValue();
        float Ry = new Float((2.f * Math.asin(1.d/Nr))/dy).floatValue();
        nh.writeVariableData("Rx", Rx);
        nh.writeVariableData("Ry", Ry);
              
        //-------------------------------------------------------------------------
        // Load the Orbit & Attitude block data
        //-------------------------------------------------------------------------
        Array orbAtt = new ArrayDouble.D1(336);
        Index iOA = orbAtt.getIndex();
        int[] oaOrigin = new int[2];
        oaOrigin[0] = 0;
        oaOrigin[1] = 0;       
        int iNav = 5;
        double d;

        logger.info(" Writing O&A data... size of nav = " + navArray.length);
        // Block 1
        for (int i=4; i<116; i++ )
            {
            if (iNav == 12)   // epoch date/time, BCD format
                {
                Long l = Long.rotateLeft(navArray[iNav++], 32);
                d = Double.longBitsToDouble(l);
                }
            else if (iNav == 13)    // *** This needs work... should it shift into previous double?
                d = navArray[iNav++];
            else if (iNav == 65)
                d = navArray[iNav++];
            else if (iNav == 14 || iNav == 61 || iNav == 63)              
                d = navArray[iNav++]/100.0d;
            else
                d = navArray[iNav++]/10000000.0d;
            
            orbAtt.setDouble(iOA.set(i), d);
       //     logger.info(i + "  " + (iNav-1) + "   " + d + "   " + navArray[iNav-1]
       //             + "   " + Integer.toHexString(navArray[iNav-1]));
            }
        
        // Block 2
        iNav = 129;
        for (int i=116; i<226; i++ )
            {
            if (iNav == 130 || iNav == 185)
                {
                d = navArray[iNav++]/100.0d;
                }
            else
                d = navArray[iNav++]/10000000.0d;
            
            orbAtt.setDouble(iOA.set(i), d);
       //     logger.info(i + "  " + (iNav-1) + "   " + d + "   " + navArray[iNav-1]);
            }
        
        // Block 3
        iNav = 257;        
        for (int i=226; i<336; i++ )
            {
            if (iNav == 258 || iNav == 313)
                {
                d = navArray[iNav++]/100.0d;
                }
            else
                d = navArray[iNav++]/10000000.0d;
            
            orbAtt.setDouble(iOA.set(i), d);
      //      logger.info(i + "  " + (iNav-1) + "   " + d + "   " + navArray[iNav-1]);
            }
       
        try
            {
            nh.writeArrayData("orbitAttitude", oaOrigin, orbAtt);
            }
        catch (Exception e)
            {
            logger.error(" * can't write O&A" + e);
            }
        
        //-------------------------------------------------------------------------
        // Load the calibration data
        //-------------------------------------------------------------------------
        int [] biasGainBandOffsets = {-99, 2, 3, 0, 1, 1};   // for bands 2,3,4,5,6

        float bias, gain, radianceToAlbedo;
        Array biasArray = new ArrayFloat.D1(numLines_nc);
        Array gainArray = new ArrayFloat.D1(numLines_nc);
        Index iBias = biasArray.getIndex();
        Index iGain = gainArray.getIndex();
        int [] origin = {0,0};
        
        if (cal != null)
            {
            logger.info(" Writing cal data... size of cal = " + cal.length);

            // Kludge: we may need to swap bytes for the cal array
            if (cal[0] > 0) ConversionUtility.swap(cal, 0, cal.length - 1);

            if (band == 1) // Vis
                {
                bias = new Double(ConversionUtility.GouldToNative(cal[0]))
                        .floatValue();
                gain = new Double(ConversionUtility.GouldToNative(cal[8]))
                        .floatValue();
                radianceToAlbedo = new Double(ConversionUtility
                        .GouldToNative(cal[24])).floatValue();
                nh.writeVariableData("radiance_to_albedo", radianceToAlbedo);
                }
            else
                // IR bands
                {
                int index = 25 + biasGainBandOffsets[band - 1];
                bias = new Double(ConversionUtility.GouldToNative(cal[index]))
                        .floatValue();

                index = 33 + biasGainBandOffsets[band - 1];
                gain = new Double(ConversionUtility.GouldToNative(cal[index]))
                        .floatValue();
                }

            logger.info(" bias is: " + bias);
            logger.info(" gain is: " + gain);

            // Write bias/gain for each line because that's how we set it up
            // in the original gvar code
            for (int line = 0; line < numLines_nc; line++)
                {
                biasArray.setFloat(iBias.set(line), bias);
                gainArray.setFloat(iGain.set(line), gain);
                }

            try
                {
                nh.writeArrayData("scalingBias", origin, biasArray);
                nh.writeArrayData("scalingGain", origin, gainArray);
                }
            catch (Exception e)
                {
                logger.info(" * can't write scaling bias/gain " + e);
                }
            }
        else
            logger.info(" ** Cal block is missing, so can't write scaling bias/gain");
        }

    }
