package gov.noaa.esrl.mcidas;

import java.io.IOException;

import edu.wisc.ssec.mcidas.AreaFileException;
import ucar.ma2.Array;
import ucar.ma2.ArrayDouble;
import ucar.ma2.ArrayFloat;
import ucar.ma2.ArrayInt;
import ucar.ma2.DataType;
import ucar.ma2.Index;

public class AreaGvarSounderWriter extends AreaDataWriter
    {

    public AreaGvarSounderWriter(AreaFileReader afr, NetCDFHandler nh) throws AreaFileException, IOException
        {
        super(afr, nh);  
        
        createSounderScanAreaArrays();
        }
    
    public void createSounderScanAreaArrays() throws IOException
        {
        numLines_nc = numLines_area;
        numElements_nc = numElements_area;
        
        int[] dims = {numBands, numLines_nc, numElements_nc};
        String[] dimNames = {"wavelength", "y", "x"}; 
        String[] attNames = {"units", "long_name", "standard_name" };
        String[] atts = {"counts", "Sounder Pixel Values", "SDATA"};
        nh.addDataVariable("sounding", dims, dimNames, DataType.SHORT, attNames, atts);
        
        int[] dims1 = {numBands, numLines_nc};
        String[] dimNames1 = {"wavelength", "y"};
        
        String[] attNames1 = {"long_name", "gvar_name", "standard_name" };
        String[] atts1 = {"Infrared Scaling Bias", "SISFB, words 3991 - 4278", "scaling_bias"};
        nh.addDataVariable("scalingBias", dims1, dimNames1, DataType.FLOAT, attNames1, atts1);

        String[] atts2 = {"Infrared Scaling Gain", "SISF1, words 4279 - 4566", "scaling_gain"};
        nh.addDataVariable("scalingGain", dims1, dimNames1, DataType.FLOAT, attNames1, atts2);
//        
//        nh.addDataVariable("lineTimeBegin", dims1, dimNames1, DataType.DOUBLE, attNames, atts);
//        nh.addDataVariable("lineTimeEnd", dims1, dimNames1, DataType.DOUBLE, attNames, atts);

        if (haveLatLons)
            {
            int[] dims2 = {numLines_nc, numElements_nc};
            String[] dimNames2 = {"y", "x"};
            String[] attNames3 = {"standard_name"};
            String[] atts3 = {"latitude"};
            nh.addDataVariable("lat", dims2, dimNames2, DataType.FLOAT, attNames3, atts3);     
        
            String[] atts4 = {"longitude"};
            nh.addDataVariable("lon", dims2, dimNames2, DataType.FLOAT, attNames3, atts4);     
            }
        }
 
    public void writeData()
        {
        ArrayInt.D3 sounding = new ArrayInt.D3(numBands, numLines_nc, numElements_nc);    
        int[] soundingOrigin = new int[3];        
        int[] llOrigin = new int[2];
        llOrigin[0] = 0;
        llOrigin[1] = 0;
        float linEles[][] = new float[2][1];
        float latLons[][] = new float[2][1];
        Array aLats = new ArrayFloat.D2(1, numElements_nc);
        Array aLons = new ArrayFloat.D2(1, numElements_nc);
        Index iLats = aLats.getIndex();
        Index iLons = aLons.getIndex();        
      
        try
            {  
            // Loop through the bands, lines, elemends
            for (int iband = 0; iband<numBands; iband++)
                for (int iline = 0; iline < numLines_nc; iline++)
                    for (int ielem = 0; ielem < numElements_nc; ielem++)
                        {    
                        // Load the values for each band
                        sounding.set(iband, iline, ielem, (short) area_data[iband][iline][ielem]);  
                        }

            // Write entire sounding array
            soundingOrigin[0] = soundingOrigin[1] = soundingOrigin[2] = 0;
            nh.writeArrayData("sounding", soundingOrigin, (Array) sounding);

            // Compute and write lat/lon arrays, if needed
            if (haveLatLons)
                {
                logger.info("write latlons:  " + haveLatLons);
                
                for (int iline = 0; iline < numLines_nc; iline++)
                    {                
                    linEles[1][0] = iline;
                    // Loop through elements in netCDF file
                    for (int ielem = 0; ielem < numElements_nc; ielem++)
                        {                     
                        linEles[0][0] = ielem;
                        latLons = nav.toLatLon(linEles);
                        aLats.setFloat(iLats.set(0, ielem), (float) latLons[0][0]);
                        aLons.setFloat(iLons.set(0, ielem), (float) latLons[1][0]);
                        
                  //      if (iline < 10 && ielem < 10)
                  //          System.out.println(iline + " " + ielem + " " + latLons[0][0] + " " + latLons[1][0]);
                        }
                   
                    llOrigin[0] = iline;
                    nh.writeArrayData(latitude, llOrigin, (Array) aLats);
                    nh.writeArrayData(longitude, llOrigin, (Array) aLons);
                    }                          
                }
            }
        catch (Exception e)
            {
            logger.error("Can't write sounder data: ", e);
            }       
        }
    
    public void writeNavAndCalVars()
        {
        String[] imcIds = {"AAAA", "BBBB", "I001", "W001", "S002", "N000", "XXXX", "YYYY", "ZZZZ"}; // for GOES-8, -9, ...
 
        final float rad2deg            = 180.0f/3.141592654f;
        final float app_earth_diameter = 2.0232f;
        final float ew_diameter        = 6378.160f;
        final float ns_diameter        = 6356.775f;
        final float Nr                 = 6.6166f; 

        int satId = (Integer) afr.adMap.get("satellite_number");
        
        //-------------------------------------------------------------------------
        // Navigation parameters
        // For GVAR nav block, see 
        // http://www.ssec.wisc.edu/mcidas/doc/prog_man/2003/formats-4.html
        //-------------------------------------------------------------------------       
        nh.writeVariableData("scan_status", navArray[2]);
        byte imc = 1;
        if ( (navArray[2] & (1 << 8)) != 0 ) imc = 0;
        nh.writeVariableData("imc_flag", imc);
        if (imc == 1 && satId <= 16)          
            nh.writeVariableData("imc_id", imcIds[satId-8]);           
        else
            nh.writeVariableData("imc_id", "    "); 
        
        byte iflip = 0;
        if ( (navArray[3] & (1 << 16)) != 0) iflip = 1;
        nh.writeVariableData("yaw_flip_flag", iflip);
        
        nh.writeVariableData("reference_longitude", navArray[5]/10000000.0f);
        nh.writeVariableData("reference_distance", navArray[6]/10000000.0f);
        nh.writeVariableData("reference_latitude", navArray[7]/10000000.0f);
        nh.writeVariableData("reference_orbit_yaw", navArray[8]/10000000.0f);
        nh.writeVariableData("reference_attitude_roll", navArray[9]/10000000.0f);
        nh.writeVariableData("reference_attitude_pitch", navArray[10]/10000000.0f);
        nh.writeVariableData("reference_attitude_yaw", navArray[11]/10000000.0f);
        nh.writeVariableData("imc_enable_time", navArray[14]/100.0f);
        nh.writeVariableData("spacecraft_comp_roll", navArray[15]/100.0f);
        nh.writeVariableData("spacecraft_comp_pitch", navArray[16]/100.0f);
        nh.writeVariableData("spacecraft_comp_yaw", navArray[16]/100.0f);       
        nh.writeVariableData("east_west_cycles", new Integer(navArray[380]).shortValue());
        nh.writeVariableData("east_west_increments", new Integer(navArray[382]).shortValue());
        nh.writeVariableData("north_south_cycles", new Integer(navArray[379]).shortValue());
        nh.writeVariableData("north_south_increments", new Integer(navArray[381]).shortValue());

        //-------------------------------------------------------------------------
        // Derived variables
        //-------------------------------------------------------------------------
        float Lop = (navArray[5]/10000000.0f) * rad2deg;
        float Lap = (navArray[7]/10000000.0f) * rad2deg;
        nh.writeVariableData("Lop", Lop);
        nh.writeVariableData("Lap", Lap);
        
        float dx = (app_earth_diameter * ew_diameter)/elemRes_area;
        float dy = (app_earth_diameter * ns_diameter)/lineRes_area;
        nh.writeVariableData("dx", dx);
        nh.writeVariableData("dy", dy);
        
        // Subpoint line/pixel are not available in Area File, 
        // so calculate from subpoint (reference) lat/lon
        float linEle[][] = new float[2][1];
        float latLon[][] = new float[2][1];
        latLon[nav.indexLat][0] = Lap;
        latLon[nav.indexLon][0] = Lop;
        linEle = nav.toLinEle(latLon);
        float xp = linEle[nav.indexEle][0] * elemRes_area + startElement_area; 
        float yp = linEle[nav.indexLine][0] * lineRes_area + startLine_area;
        nh.writeVariableData("Xp", new Float(xp).shortValue());
        nh.writeVariableData("Yp", new Float(yp).shortValue());    
        
        float orientation = 0.f;   // not sure how to derive this from Area file
        nh.writeVariableData("orientation", orientation);
        nh.writeVariableData("Nr", Nr);
        
        float Rx = new Float((2.f * Math.asin(1.d/Nr))/dx).floatValue();
        float Ry = new Float((2.f * Math.asin(1.d/Nr))/dy).floatValue();
        nh.writeVariableData("Rx", Rx);
        nh.writeVariableData("Ry", Ry);
              
        //-------------------------------------------------------------------------
        // Load the Orbit & Attitude block data
        //-------------------------------------------------------------------------
        Array orbAtt = new ArrayDouble.D1(336);
        Index iOA = orbAtt.getIndex();
        int[] oaOrigin = new int[2];
        oaOrigin[0] = 0;
        oaOrigin[1] = 0;       
        int iNav = 5;
        double d;

        logger.info(" Writing O&A data... size of nav = " + navArray.length);
        // Block 1
        for (int i=4; i<116; i++ )
            {
            if (iNav == 12)   // epoch date/time, BCD format
                {
                Long l = Long.rotateLeft(navArray[iNav++], 32);
                d = Double.longBitsToDouble(l);
                }
            else if (iNav == 13)    // *** This needs work... should it shift into previous double?
                d = navArray[iNav++];
            else if (iNav == 65)
                d = navArray[iNav++];
            else if (iNav == 14 || iNav == 61 || iNav == 63)              
                d = navArray[iNav++]/100.0d;
            else
                d = navArray[iNav++]/10000000.0d;
            
            orbAtt.setDouble(iOA.set(i), d);
       //     logger.info(i + "  " + (iNav-1) + "   " + d + "   " + navArray[iNav-1]
       //             + "   " + Integer.toHexString(navArray[iNav-1]));
            }
        
        // Block 2
        iNav = 129;
        for (int i=116; i<226; i++ )
            {
            if (iNav == 130 || iNav == 185)
                {
                d = navArray[iNav++]/100.0d;
                }
            else
                d = navArray[iNav++]/10000000.0d;
            
            orbAtt.setDouble(iOA.set(i), d);
       //     logger.info(i + "  " + (iNav-1) + "   " + d + "   " + navArray[iNav-1]);
            }
        
        // Block 3
        iNav = 257;        
        for (int i=226; i<336; i++ )
            {
            if (iNav == 258 || iNav == 313)
                {
                d = navArray[iNav++]/100.0d;
                }
            else
                d = navArray[iNav++]/10000000.0d;
            
            orbAtt.setDouble(iOA.set(i), d);
      //      logger.info(i + "  " + (iNav-1) + "   " + d + "   " + navArray[iNav-1]);
            }
       
        try
            {
            nh.writeArrayData("orbitAttitude", oaOrigin, orbAtt);
            }
        catch (Exception e)
            {
            logger.error(" * can't write O&A" + e);
            }
        
        //-------------------------------------------------------------------------
        // Load the calibration data
        //-------------------------------------------------------------------------
        float[] bias = new float[18];
        float[] gain = new float[18];
        Array aBias = new ArrayFloat.D2(numBands, numLines_nc);
        Array aGain = new ArrayFloat.D2(numBands, numLines_nc);
        Index iBias = aBias.getIndex();
        Index iGain = aGain.getIndex();
        int [] origin = {0,0};
        
        logger.info(" Writing cal bias/gain data... size of cal = " + cal.length);

        // Kludge: need to swap bytes for the cal array if we're reading a physical file
        //  So, just check if the first bias value is reasonable
        if (ConversionUtility.GouldToNative(cal[13]) < 1.0f)
            {
            logger.info(" Swapping cal bytes");
            ConversionUtility.swap(cal, 0, cal.length-1);
            }

        // Bias for each band starts at cal Word 14
        // Gain for each band starts at cal Word 32
        for (int iband=0; iband<18; iband++)
            {
            bias[iband] = new Double(ConversionUtility.GouldToNative(cal[13 + iband])).floatValue();
            gain[iband] = new Double(ConversionUtility.GouldToNative(cal[31 + iband])).floatValue();
            //logger.info(" bias for band " + iband + " is: " + bias[iband]);
            //logger.info(" gain for band " + iband + " is: " + gain[iband]);

            // Write bias/gain for each line because that's how we set it up
            // in the original gvar code
            for(int line=0; line<numLines_nc; line++)
                {
                aBias.setFloat(iBias.set(iband,line), bias[iband]);
                aGain.setFloat(iGain.set(iband,line), gain[iband]);
                }

            try
                {
                nh.writeArrayData("scalingBias", origin, aBias);
                nh.writeArrayData("scalingGain", origin, aGain);
                }
            catch (Exception e)
                {
                logger.info(" * can't write scaling bias/gain " + e);
                }  
            }
        }
   
    
    public void writeStandardVars()
        {
        // Fill in some standard variables
        nh.writeVariableData("reference_time", (Object) new Double(imageTime)
                .doubleValue());
        nh.writeVariableData("ny", (Object) numLines_nc);
        nh.writeVariableData("nx", (Object) numElements_nc);

        nh.writeVariableData("start_line", (Object) startLine_area);
        nh.writeVariableData("start_element", (Object) startElement_area);

        int endLine = startLine_area + numLines_area - 1;
        int endElem = startElement_area + numElements_area - 1;
        nh.writeVariableData("end_line", (Object) endLine);
        nh.writeVariableData("end_element", (Object) endElem);
        }
    }
