package gov.noaa.esrl.mcidas;

import gnu.getopt.Getopt;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

public class McArea2NetCDF
    {
    private static Logger logger;
    private static boolean verbose = true;
    private String cmdLine = null;

    private static String areaFile;
    private static String ncml;
    private static String outPath = null;
    private static String logPath = null;
    private static String procName = null;
    private static String extension = ".nc";         // default
    private static boolean replaceExisting = false;  // default: don't clobber existing file
   
    public McArea2NetCDF(String[] args) throws Exception
        {         
        handleArgs(args);      
        logger = new MyLogger(logPath, verbose).logger;
        logger.info(cmdLine);

        // Check for required args
        if (areaFile == null || outPath == null || ncml == null )
            {
            System.out.println(usage("missing arguments"));
            throw new McAreaException("Exiting");
            }     
        
        AreaFileReader afr = null; 
        NetCDFHandler nh = null;       
        AreaDataWriter adw = null;
        try
            {
            // Get the McIDAS Area File
            afr = new AreaFileReader(areaFile);
            // afr.dumpAreaDirectoryMap(); 
            
            int startTime = afr.getStartTime();
            logger.info("Area file start time: " + startTime + " - " + 
                    new SimpleDateFormat("yyyy-MM-dd HH:mm:ss").format(new Date(startTime * 1000L)));
            
            long now = Calendar.getInstance().getTimeInMillis();            
            int secsSinceImageStart = (int) (now/1000L) - startTime;
            logger.info("Secs since image start time:  " + secsSinceImageStart);
            
            // Initialize the netCDF file
            nh = new NetCDFHandler(ncml, outPath, (Date) afr.adMap.get("start_time"), extension, replaceExisting);
            logger.info("NetCDF filename: " + nh.getFilename());          

            // Get data type from Area Directory info       
            String areaType = (String) afr.adMap.get("source_type")
                  + "/" + (String) afr.adMap.get("sensor_type");
            
            String navType = "unknown";
            if (afr.getAreaNavigation() != null)
                navType = afr.getAreaNavigation().toString();
            logger.info("Source/sensor type: " + areaType + "   Nav type: " + navType);
           
            // Construct a datatype-specific writer
            if (Pattern.matches("GVAR.*imager", areaType) && navType.equals("GVAR"))
                {
                logger.info("Have GVAR imager data");
                adw = new AreaGvarImagerWriter(afr, nh);
                }  
            else if (Pattern.matches("GVAR.*sounder", areaType) && navType.equals("GVAR"))
                {
                logger.info("Have GVAR sounder data");
                adw = new AreaGvarSounderWriter(afr, nh);
                }
            else
                {
                logger.info("Have non-GVAR image data");
                adw = new AreaImageWriter(afr, nh);
                }
            
            // Write the data into the netCDF
            if (verbose) adw.dumpAreaParameters();  
            adw.writeData();
            adw.writeNavAndCalVars();
            adw.writeStandardVars();             
            }
        catch (Exception e)
            {
            e.printStackTrace();
            throw new McAreaException("can't process Area File");
            }      
            
        // Wrap-up the netCDF file
        nh.writeGlobalAttribute("created", new Date().toString());
        nh.writeGlobalAttribute("areaFile", areaFile);
        nh.closeFile();
        
        logger.info("Finished file " + nh.getFilename());
        
        // Emit a standard text notification, if necessary
        if (procName != null) textNotify(nh.getFilename());
        }
    
    private String usage(String msg)
        {
        return new String(
                msg
                   + "\nUsage : " + this.getClass().getSimpleName() + " "
                   + "-a <area file> -o <output path> -n <ncml file> (or -x <xml file>) [-e <extension>] [-r(eplace existing)] [-p <process name>] [-l <log path>] [-v(erbose)]");
        }
     
    private void handleArgs(String[] args)
        {
        cmdLine = this.getClass().getSimpleName();
        for (int i=0; i<args.length; i++)            
            cmdLine = cmdLine + " " + args[i];
        
        // Parse the arguments...
        Getopt opt = new Getopt(McArea2NetCDF.class.getName(), args,
                "a:o:e:n:x:l:p:rvh");
        opt.setOpterr(false);
        int c;
        while ((c = opt.getopt()) != -1)
            {
            switch (c)
                {
                case 'a':
                    areaFile = opt.getOptarg();
                break;
                
                case 'o':
                    outPath = opt.getOptarg();
                break;

                case 'e':
                    extension = opt.getOptarg();
                break;
               
                case 'n':
                case 'x':
                    ncml = opt.getOptarg();
                break;

                case 'l':
                    logPath = opt.getOptarg();
                break;
 
                case 'p':
                    procName = opt.getOptarg();
                break;
                
                case 'r':
                    replaceExisting = true;
                break;

                case 'v':
                    verbose = true;
                break;

                case 'h':
                    System.out.println(usage(""));
                    AreaFileReader.listMcIDASNames();
                    System.exit(0);
                    
                case '?':
                default:
                    // Throw an exception with usage
                    System.out.println(usage("Invalid arg...\n" + cmdLine));
                    System.exit(-1);
                }
            }
        }

    public void textNotify(String outFile)
        {
        String logArg = "";
        if (logPath != null) logArg = " -l " + logPath;
        String ntyCmd = "textNotify.pl -F " + outFile + " -k " + procName
                + " -m MMXX54 -v" + logArg;
        logger.info(ntyCmd);
        try
            {
            Process p = Runtime.getRuntime().exec(ntyCmd);
            try
                {
                p.waitFor();
                }
            catch (InterruptedException e)
                {
                logger.error("Error waiting for textNotify.pl:", e);
                }
            }
        catch (IOException e)
            {
            logger.error("Can't call textNotify.pl:", e);
            }
        }

    public static void main(String[] args)
        {       
        outPath = "/tmp";
        replaceExisting = true;
        verbose = true;

        
        String test = "";
         // test = "msg_wv";
         // test = "goes-west";
         // test = "ger";
         // test = "goes12";
         // test = "awc_ir";
         // test = "cira";
         // test = "mtsat";
         // test = "meteosat";
         // test = "goes11_sdr";
         // test = "goes12_hawaii";
         // test = "goes12_katrina";
         // test = "fy";
         // test = "him8";

         if (test.equals("goes12"))
             {
             //  areaFile = "/dsg-scratch/bob/data/mcidas/AREA3004";
             // areaFile = "/dsg-scratch/bob/data/mcidas/AREA3008";
             //areaFile = "/dsg-scratch/bob/data/sat/dan/goes12.2008.143.174013.BAND_01";
             areaFile = "/dsg-scratch/bob/data/mcidas/AREA3002_gvar_2837-9277";
             ncml = "/home/lipschut/netcdf/area2nc_gvar_ir_4km_fsl-conus.xml";
             extension = "_11u";
             }
         else if (test.equals("goes12_hawaii"))
             {
             areaFile = "/dsg-scratch/bob/data/sat/200912031815.ui4";
          //   areaFile = "/dsg-scratch/bob/data/sat/G12-200912172145.gstn.ir4.conus.ara";
             ncml = "/home/lipschut/netcdf/area2nc_gvar_ir_4km_fsl-conus.xml";
          //   ncml = "/home/lipschut/netcdf/area2nc_gvar_vis_2km_aoml.xml";
             extension = "_11u";
             }
         else if (test.equals("goes12_katrina"))
             {
           //  areaFile = "/dsg-scratch/bob/data/sat/goes12/area/goes12.2005.239.174514.BAND_04";
          //   ncml = "/home/lipschut/netcdf/area2nc_gvar_ir_4km_fsl-conus.xml";
          //   extension = "_11u";
             
             areaFile = "/dsg-scratch/bob/data/sat/goes12/area/goes12.2005.239.174514.BAND_01";
             ncml = "/home/lipschut/netcdf/area2nc_gvar_vis_2km_aoml.xml";
             extension = "_vis";
             }
       else if (test.equals("goes-west")) 
             {
             areaFile = "/dsg-scratch/bob/data/mcidas/AREA3008";
           //  areaFile = "/dsg-scratch/bob/data/mcidas/AREA2097";
             ncml = "/home/lipschut/netcdf/area2nc_gvar_ir_4km_fsl-pacus.xml";
             extension = "_11u";
             }
         else if (test.equals("goes11_sdr"))
             {
             areaFile = "/dsg-scratch/bob/data/mcidas/AREA3004_sdr";
             ncml = "/home/lipschut/netcdf/area2nc_gvar_sounder.xml";
             extension = "_sdr";
             }
         else if (test.equals("awc_ir"))
             {     
             areaFile = "/dsg-scratch/bob/data/mcidas/WRLD-IR10KM_20080819_1645";
             // ncml = "/home/lipschut/netcdf/area2nc_awc_ir10km.xml";
             ncml = "/home/lipschut/netcdf/area2nc_awc_world_10km_ir.ncml";
             }
         else if (test.equals("ger"))
             {
             areaFile =
                 //    "adde://SATEPSDIST2E.NESDIS.NOAA.GOV/imagedata?group=GER&descr=GECONS04I4.0&mag=-4&doc=yes";
                 //    "adde://140.90.213.56/imagedata?group=GER&descr=GECONS04I4.0&mag=-4&doc=yes";
                 "adde://geodist2e.espc.nesdis.noaa.gov/imagedata?group=GER&descr=GECONS04I4&cal=raw&unit=raw&size=all&debug=true&track=yes";            
             ncml = "/home/lipschut/netcdf/area2nc_gvar_ir_4km_fsl-conus.xml";
             }
         else if (test.equals("gini_east"))
            {
            areaFile = "adde://ADDE.UCAR.EDU/imagedata?group=GINIEAST&descr=GE4KIR&size=all";
            ncml = "/home/lipschut/netcdf/area2nc_gini_ge_4kir.xml"; 
            }
         else if (test.equals("cira"))
             {
             areaFile = "/home/lipschut/mcidas/cira/cd/Test_Data/2008119144514i12.c04";
             ncml = "/home/lipschut/netcdf/area2nc_gvar_ir_4km_fsl-pacus.xml"; 
             }
         else if (test.equals("mtsat"))
             {
             areaFile = //"/dsg-scratch/bob/data/mcidas/AREA3002_0810311530_mtsat";
             //   "adde://SATEPSANONE.NESDIS.NOAA.GOV/imagedata?group=PUB&descr=MTGLOB04I2&unit=raw&size=all&&debug=true&doc=yes&mag=-2";
             // ncml = "/home/lipschut/netcdf/area2nc_mtsat_fulldisk.xml"; 
                "adde://SATEPSANONE.NESDIS.NOAA.GOV/imagedata?group=PUB&descr=MTGLOB04I2&unit=temp&size=all&&debug=true&doc=yes&mag=-2";
             ncml = "/home/lipschut/netcdf/area2nc_mtsat_temp_fulldisk.xml"; 
             extension = "_10p8u";
             procName = "test.test";
             }
         else if (test.equals("him8"))
             {
             areaFile = // "/dsg-scratch/bob/data/mcidas/HMGLOB04I13.area";
                     "adde://geodist7e.espc.nesdis.noaa.gov/imagedata?group=HIM&descr=HMGLOB04I13&unit=raw&size=all&&debug=true&doc=yes&mag=-2";
             ncml = "/home/lipschut/netcdf/area2nc_him8_ir_8km_full-disk.xml"; 
             extension = "_10p4u.nc";
             procName = "test.test";
             }
        else if (test.equals("msg_wv"))
             {
             areaFile = "adde://SATEPSDIST6E.NESDIS.NOAA.GOV/imagedata?group=MSG&descr=MSGGLOB01V&unit=raw&size=all&&debug=true&doc=yes&mag=-3";
          //   areaFile = "adde://SATEPSDIST6E.NESDIS.NOAA.GOV/imagedata?group=MSG&descr=MSGLOB05WV&unit=temp&size=all&&debug=true&doc=yes&mag=-3";

             ncml = "/home/lipschut/netcdf/msg_9km_wv.xml"; 
             extension = ".nc";
             procName = "test.test";
             }
         else if (test.equals("meteosat"))
             {
             areaFile = "/dsg-scratch/bob/data/mcidas/AREA3001_msg";            
             ncml = "/scratch/lipschut/rt/etc/h1/area2nc_meteosat_ir_9km_full-disk.xml"; 
             extension = "_10p8u";
             System.out.println("meteosat");
             }
         else if (test.equals("fy"))
             {
             areaFile = 
                "adde://SATEPSANONE.NESDIS.NOAA.GOV/imagedata?group=PUB&descr=FY2D-CH2&unit=temp&size=all&&debug=true&doc=yes&mag=-2";            
             ncml = "/scratch/lipschut/rt/etc/dc/area2nc_fy_ir_10km_full-disk.xml"; 
             extension = "_10p8.nc";
             System.out.println("FY2D");
             }

         else
             {
             replaceExisting = false;
             verbose = false;
             }

         McArea2NetCDF m2n = null;
         try
            {          
            // Do the work
            m2n = new McArea2NetCDF(args);
            }
         catch (Exception ex)
            {
            logger.error("Error exit", ex);
            System.exit(-1);
            }
        
        logger.info("done");
        System.exit(0);
        }
    }
