package gov.noaa.esrl.mcidas;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.TreeMap;

import org.apache.log4j.Logger;

import ucar.ma2.Array;
import ucar.ma2.ArrayByte;
import ucar.ma2.ArrayChar;
import ucar.ma2.ArrayDouble;
import ucar.ma2.ArrayFloat;
import ucar.ma2.ArrayInt;
import ucar.ma2.ArrayLong;
import ucar.ma2.ArrayShort;
import ucar.ma2.DataType;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriteable;
import ucar.nc2.Variable;
import ucar.nc2.ncml.NcMLReader;


public class NetCDFHandler
    {
    private Logger logger = null;
    private boolean verbose = false;
   
    private NetcdfFileWriteable ncOutW = null;
    private String _ncmlFile;
    private String _outFile;
    private boolean _replaceExisting = false;
    
    public int numElements_nc = 0;
    public int numLines_nc = 0;
    public int startLine_nc = 0;
    public int startElement_nc= 0;
    public int lineRes_nc = 1;
    public int elemRes_nc = 1;
    public int[] startLine_gvar = {0,0,0,0,0,0,0,0,0};
    public int[] startElement_gvar = {0,0,0,0,0,0,0,0,0};
   
    public TreeMap<String,Variable> stdNameMap;
    public TreeMap<String, Variable> mcNameMap;

    public String latName = "lat";    // default
    public String lonName = "lon";    // default
      
    public NetCDFHandler(String ncmlFile, String outFile, 
            boolean replaceExisiting) throws Exception
        {
        _ncmlFile = ncmlFile;
        _outFile = outFile;
        _replaceExisting = replaceExisiting;

        logger = MyLogger.logger;
        verbose = MyLogger.verbose;
       
        openFile();
        }
    
    public NetCDFHandler(String ncmlFile, String outPath, Date startTime, String extension, 
            boolean replaceExisiting) throws Exception
        {
        _ncmlFile = ncmlFile;
        _replaceExisting = replaceExisiting;
        _outFile = outPath + "/" + dateToFilename(startTime, extension);

        logger = MyLogger.logger;
        verbose = MyLogger.verbose;
 
        openFile();
        }
   
    private void openFile() throws McAreaException 
        {
        // Bail if this isn't a new file and we're not wanting to "replace
        // existing"
        if (new File(_outFile).exists() && ! _replaceExisting )
            {
            throw new McAreaException("File already exists... use '-r' to replace... exiting");
            }
        try
            {
            // Create the output file (this is like doing ncgen)
            NcMLReader.writeNcMLToFile(_ncmlFile, _outFile);            

            // Open for writing the file we just created
            ncOutW = NetcdfFileWriteable.openExisting(_outFile);
            
            buildNameMaps();
            loadXYDims();
            loadStartLineElement();
            loadResolution();
            }
        catch (IOException ioe)
            {
            ioe.printStackTrace();
            throw new McAreaException("can't open " + _outFile);
            }       
        }
    
    public void closeFile()
        {
        try
            {
       //     ncOutW.sync();
            ncOutW.close();
            }
        catch (IOException e)
            {
            logger.error("Can't close file: ", e);
            }
        }
   
    private void buildNameMaps()
        {
        String stdName, mcName;
        Variable v;
        Attribute a;
        stdNameMap = new TreeMap<String,Variable>();
        mcNameMap = new TreeMap<String,Variable>();
       
        // Construct maps for standard_name and mcidas_name variables
        List vars = ncOutW.getVariables();
        for (int i = 0; i < vars.size(); i++)
            {
            v = (Variable) vars.get(i);
            
            // If we have a "standard_name" attribute, add to stdName map
            a = v.findAttribute("standard_name");
            if (a != null) 
                {
                stdName = a.getStringValue();
                stdNameMap.put(stdName, v);
                }

            // If we have a "mcidas_name" attribute, add to mcName map
            a = v.findAttribute("mcidas_name");
            if (a != null) 
                {
                mcName = a.getStringValue();
                mcNameMap.put(mcName, v);
                }
            }
        }
    
    private void loadXYDims()
        {       
        // Get line/element (y,x) dimensions
        Dimension dX = ncOutW.findDimension("x");
        Dimension dY = ncOutW.findDimension("y");
        if (dX != null && dY != null)
            {
            numElements_nc = dX.getLength();
            numLines_nc = dY.getLength();
            }
        }
    
    public int[] getXYDims()
        {
        int[] xy = { numElements_nc, numLines_nc };
        return (xy);
        }
    
    public boolean haveLatLons()
        {
        boolean haveLatLons = false;
        // Find if lat/lon are included
        if (stdNameMap.containsKey("latitude") && 
                stdNameMap.containsKey("longitude")) 
            {
            haveLatLons = true; 
            latName = stdNameMap.get("latitude").getName();
            lonName = stdNameMap.get("longitude").getName();
            } 
        
        // Force dynamic lat/lon's with this global attribute (needed for GVAR sounder)
        Attribute writeLatLons = ncOutW.findGlobalAttribute("createLatLons");
        if (writeLatLons != null) haveLatLons = true;
        
        return haveLatLons;
        }

    private void loadStartLineElement()
        {
        // Get starting line/element from global attributes
        // *** Note:
        // For GVAR, may specify starting coords for multiple
        // satellites (GOES-8, -9, -10, -11, -12,...)
        Attribute aStartLine = ncOutW.findGlobalAttribute("startLine");
        Attribute aStartElem = ncOutW.findGlobalAttribute("startElem");
        if (aStartLine != null)
            startLine_nc = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_nc = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g08");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g08");
        if (aStartLine != null)
            startLine_gvar[0] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[0] = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g09");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g09");
        if (aStartLine != null)
            startLine_gvar[1] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[1] = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g10");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g10");
        if (aStartLine != null)
            startLine_gvar[2] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[2] = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g11");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g11");
        if (aStartLine != null)
            startLine_gvar[3] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[3] = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g12");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g12");
        if (aStartLine != null)
            startLine_gvar[4] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[4] = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g13");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g13");
        if (aStartLine != null)
            startLine_gvar[5] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[5] = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g14");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g14");
        if (aStartLine != null)
            startLine_gvar[6] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[6] = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g15");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g15");
        if (aStartLine != null)
            startLine_gvar[7] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[7] = aStartElem.getNumericValue().intValue();

        aStartLine = ncOutW.findGlobalAttribute("startLine_g16");
        aStartElem = ncOutW.findGlobalAttribute("startElem_g16");
        if (aStartLine != null)
            startLine_gvar[8] = aStartLine.getNumericValue().intValue();
        if (aStartElem != null)
            startElement_gvar[8] = aStartElem.getNumericValue().intValue();
        }
    
    private void loadResolution()
        {      
        // Get line/element resolution from global attributes
        Attribute aLineRes = ncOutW.findGlobalAttribute("lineRes");
        Attribute aElemRes = ncOutW.findGlobalAttribute("elemRes");
        if (aLineRes != null)
            {
            lineRes_nc = aLineRes.getNumericValue().intValue();
            logger.info("lineRes_nc from attribute: " + lineRes_nc);
            }
        if (aElemRes != null)
            {
            elemRes_nc = aElemRes.getNumericValue().intValue();
            logger.info("elemRes_nc from attribute: " + elemRes_nc);
            }
        }
    
    public void dumpMap(TreeMap<String, Variable> map)
        {
        Iterator iterator = map.keySet().iterator();
        System.out.println("=== names from map: ");
        while (iterator.hasNext())
            {
            System.out.println(iterator.next().toString());
            }
        System.out.println("===");
        }
    
    /*
     * writeVariableData - use variable associated with standard_name, if
     * defined
     */
    public void writeVariableData(String name, Object value)
        {
        String n = name;
       
        // use variable associated with standard_name, if it's defined
        if (stdNameMap.containsKey(name))
            {
            n = stdNameMap.get(name).getName();
            }
     
        try
            {
            DataType t = ncOutW.findVariable(n).getDataType();
            if (verbose) 
                logger.info(" Writing \"" + n + "\"   value is: " + value + 
                    "   type is: " + t.toString());
            
            if (t == DataType.BYTE)
                {
                ArrayByte.D0 val = new ArrayByte.D0();
                val.set((Byte) value);
                ncOutW.write(n, val);
                }
            else if (t == DataType.INT)
                {
                ArrayInt.D0 val = new ArrayInt.D0();
                val.set((Integer) value);
                ncOutW.write(n, val);
                }
            else if (t == DataType.SHORT)
                {
                ArrayShort.D0 val = new ArrayShort.D0();
                // Check if we need to convert int to short, as for a few
                // GVAR variables like x_resolution
                if (value instanceof java.lang.Integer)
                    {
                    short svalue = new Integer((Integer) value).shortValue();
                    val.set((Short) svalue);
                    }
                else                       
                    val.set((Short) value);
                ncOutW.write(n, val);
                }
            else if (t == DataType.FLOAT)
                {
                ArrayFloat.D0 val = new ArrayFloat.D0();
                val.set((Float) value);
                ncOutW.write(n, val);
                }
            else if (t == DataType.LONG)
                {
                ArrayLong.D0 val = new ArrayLong.D0();
                val.set((Long) value);
                ncOutW.write(n, val);
                }
            else if (t == DataType.DOUBLE)
                {
                ArrayDouble.D0 val = new ArrayDouble.D0();
                val.set((Double) value);
                ncOutW.write(n, val);
                }
            else if (t == DataType.CHAR)
                {
                ArrayChar.D1 val = new ArrayChar.D1(((String) value).length()); 
                val.setString((String) value);
                ncOutW.write(n, val);
                }
            else
                {
                logger.error(" ** " + n + " -- data type for variable not handled");
                }
            }
        catch (Exception e)
            {
            logger.info(" ** No variable for " + n + " (" + value + ")" + e);
            }
        }

    public void writeGlobalAttribute(String attribute, Object value)
        {
        try
            {     
            Attribute a = ncOutW.findGlobalAttribute(attribute);
            
            if (a != null)
                {
                DataType t = a.getDataType();
                if (t == DataType.INT)
                    {
                    Attribute aa = new Attribute (attribute, (Integer) value);
                    ncOutW.updateAttribute(null, aa);
                    }               
                else if (t == DataType.STRING)
                    {
                    Attribute aa = new Attribute (attribute, (String) value);
                    ncOutW.updateAttribute(null, aa);
                    }              
                logger.info(" Writing attribute  \"" + attribute + "\"  " + value);
                }
            }
        catch (Exception e)
            {
            logger.error("Can't update attribute " + attribute + " (" + value + "): " + e);
            }
        }
 
    public void writeArrayData(String name, int[] origin, Array data) throws IOException, InvalidRangeException
        {
        String n = name;
        
        // use variable associated with standard_name, if it's defined
        if (stdNameMap.containsKey(name))
            {
            n = stdNameMap.get(name).getName();
            }
        
        ncOutW.write(n, origin, data);
        }
    
    public void addDataVariable(String name, int[] dims, String[] dimNames,  
            DataType dataType,
            String[] attNames,
            Object[] atts) throws IOException
        {
        logger.info("Adding dims and variable for " + name);

        // Temporarily put netCDF file into define mode
        ncOutW.setRedefineMode(true); 
        String dimsString = ""; 
        
        // Create dimension, if necessary
        for (int i=0; i<dims.length; i++)
            {
            if (ncOutW.findDimension(dimNames[i]) == null)
                {
                ncOutW.addDimension(dimNames[i], dims[i]);
                logger.info("Added new dim " + dimNames[i] + "  length: " + dims[i]);
                }
            else
                {
                logger.info("Dim " + dimNames[i] + " already exists");
                }
            dimsString += dimNames[i] + " ";
            }
 
        ncOutW.addVariable(name, dataType, dimsString);
        
        for (int i=0; i<atts.length; i++)
            {
            ncOutW.addVariableAttribute(name, attNames[i], (String) atts[i]);
            }
        
        // Done adding stuff, so get out of define mode
        ncOutW.setRedefineMode(false);
        }
    
    public String dateToFilename(Date imgTime, String ext) throws Exception
        {
        SimpleDateFormat sdf = new SimpleDateFormat("yyDDDHHmm");

        return sdf.format(imgTime) + ext;
        }
    
    public String getFilename()
        {
        return _outFile;
        }
      
    }
